@@ -39,6 +39,7 @@
 
 #include "gstffmpeg.h"
 #include "gstffmpegcodecmap.h"
+#include "gstffmpegutils.h"
 #include "gstffmpegpipe.h"
 
 typedef struct _GstFFMpegDemux GstFFMpegDemux;
@@ -55,6 +56,8 @@ struct _GstFFStream
   gboolean discont;
   gboolean eos;
   GstFlowReturn last_flow;
+
+  GstTagList *tags;             /* stream tags */
 };
 
 struct _GstFFMpegDemux
@@ -89,18 +92,15 @@ struct _GstFFMpegDemux
   /* cached seek in READY */
   GstEvent *seek_event;
 
+  /* cached upstream events */
+  GList *cached_events;
+
   /* push mode data */
   GstFFMpegPipe ffpipe;
   GstTask *task;
   GStaticRecMutex *task_lock;
 };
 
-typedef struct _GstFFMpegDemuxClassParams
-{
-  AVInputFormat *in_plugin;
-  GstCaps *sinkcaps, *videosrccaps, *audiosrccaps;
-} GstFFMpegDemuxClassParams;
-
 typedef struct _GstFFMpegDemuxClass GstFFMpegDemuxClass;
 
 struct _GstFFMpegDemuxClass
@@ -180,20 +180,29 @@ static void
 gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass)
 {
   GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
-  GstFFMpegDemuxClassParams *params;
+  AVInputFormat *in_plugin;
+  gchar *p, *name;
   GstElementDetails details;
+  GstCaps *sinkcaps;
   GstPadTemplate *sinktempl, *audiosrctempl, *videosrctempl;
 
-  params = (GstFFMpegDemuxClassParams *)
+  in_plugin = (AVInputFormat *)
       g_type_get_qdata (G_OBJECT_CLASS_TYPE (klass), GST_FFDEMUX_PARAMS_QDATA);
-  g_assert (params != NULL);
+  g_assert (in_plugin != NULL);
+
+  p = name = g_strdup (in_plugin->name);
+  while (*p) {
+    if (*p == '.' || *p == ',')
+      *p = '_';
+    p++;
+  }
 
   /* construct the element details struct */
-  details.longname = g_strdup_printf ("FFMPEG %s demuxer",
-      params->in_plugin->long_name);
+  details.longname = g_strdup_printf ("FFmpeg %s demuxer",
+      in_plugin->long_name);
   details.klass = "Codec/Demuxer";
-  details.description = g_strdup_printf ("FFMPEG %s demuxer",
-      params->in_plugin->long_name);
+  details.description = g_strdup_printf ("FFmpeg %s demuxer",
+      in_plugin->long_name);
   details.author = "Wim Taymans <wim@fluendo.com>, "
       "Ronald Bultje <rbultje@ronald.bitfreak.net>, "
       "Edward Hervey <bilboed@bilboed.com>";
@@ -202,22 +211,19 @@ gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass)
   g_free (details.description);
 
   /* pad templates */
-  gst_caps_ref (params->sinkcaps);
+  sinkcaps = gst_ffmpeg_formatid_to_caps (name);
   sinktempl = gst_pad_template_new ("sink",
-      GST_PAD_SINK, GST_PAD_ALWAYS, params->sinkcaps);
+      GST_PAD_SINK, GST_PAD_ALWAYS, sinkcaps);
   videosrctempl = gst_pad_template_new ("video_%02d",
-      GST_PAD_SRC, GST_PAD_SOMETIMES, params->videosrccaps);
+      GST_PAD_SRC, GST_PAD_SOMETIMES, GST_CAPS_ANY);
   audiosrctempl = gst_pad_template_new ("audio_%02d",
-      GST_PAD_SRC, GST_PAD_SOMETIMES, params->audiosrccaps);
-
-  params->videosrccaps = NULL;
-  params->audiosrccaps = NULL;
+      GST_PAD_SRC, GST_PAD_SOMETIMES, GST_CAPS_ANY);
 
   gst_element_class_add_pad_template (element_class, videosrctempl);
   gst_element_class_add_pad_template (element_class, audiosrctempl);
   gst_element_class_add_pad_template (element_class, sinktempl);
 
-  klass->in_plugin = params->in_plugin;
+  klass->in_plugin = in_plugin;
   klass->videosrctempl = videosrctempl;
   klass->audiosrctempl = audiosrctempl;
   klass->sinktempl = sinktempl;
@@ -324,8 +330,11 @@ gst_ffmpegdemux_close (GstFFMpegDemux * demux)
     GstFFStream *stream;
 
     stream = demux->streams[n];
-    if (stream && stream->pad) {
-      gst_element_remove_pad (GST_ELEMENT (demux), stream->pad);
+    if (stream) {
+      if (stream->pad)
+        gst_element_remove_pad (GST_ELEMENT (demux), stream->pad);
+      if (stream->tags)
+        gst_tag_list_free (stream->tags);
       g_free (stream);
     }
     demux->streams[n] = NULL;
@@ -474,8 +483,9 @@ gst_ffmpegdemux_do_seek (GstFFMpegDemux * demux, GstSegment * segment)
   }
 
   GST_DEBUG_OBJECT (demux,
-      "About to call av_seek_frame (context, %d, %lld, 0) for time %"
-      GST_TIME_FORMAT, index, fftarget, GST_TIME_ARGS (target));
+      "About to call av_seek_frame (context, %d, %" G_GINT64_FORMAT
+      ", 0) for time %" GST_TIME_FORMAT, index, fftarget,
+      GST_TIME_ARGS (target));
 
   if ((seekret =
           av_seek_frame (demux->context, index, fftarget,
@@ -956,6 +966,7 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
   stream->avstream = avstream;
   stream->last_ts = GST_CLOCK_TIME_NONE;
   stream->last_flow = GST_FLOW_OK;
+  stream->tags = NULL;
 
   switch (ctx->codec_type) {
     case CODEC_TYPE_VIDEO:
@@ -1017,12 +1028,11 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
 
   /* metadata */
   if ((codec = gst_ffmpeg_get_codecid_longname (ctx->codec_id))) {
-    GstTagList *list = gst_tag_list_new ();
+    stream->tags = gst_tag_list_new ();
 
-    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
+    gst_tag_list_add (stream->tags, GST_TAG_MERGE_REPLACE,
         (ctx->codec_type == CODEC_TYPE_VIDEO) ?
         GST_TAG_VIDEO_CODEC : GST_TAG_AUDIO_CODEC, codec, NULL);
-    gst_element_found_tags_for_pad (GST_ELEMENT (demux), pad, list);
   }
 
   return stream;
@@ -1122,9 +1132,10 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   GstFFMpegDemuxClass *oclass =
       (GstFFMpegDemuxClass *) G_OBJECT_GET_CLASS (demux);
   gchar *location;
-  gint res, n_streams;
+  gint res, n_streams, i;
   GstTagList *tags;
   GstEvent *event;
+  GList *cached_events;
 
   /* to be sure... */
   gst_ffmpegdemux_close (demux);
@@ -1154,19 +1165,12 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
 
   /* open_input_file() automatically reads the header. We can now map each
    * created AVStream to a GstPad to make GStreamer handle it. */
-  for (res = 0; res < n_streams; res++) {
-    gst_ffmpegdemux_get_stream (demux, demux->context->streams[res]);
+  for (i = 0; i < n_streams; i++) {
+    gst_ffmpegdemux_get_stream (demux, demux->context->streams[i]);
   }
 
   gst_element_no_more_pads (GST_ELEMENT (demux));
 
-  /* grab the tags */
-  tags = gst_ffmpegdemux_read_tags (demux);
-  if (tags) {
-    gst_element_post_message (GST_ELEMENT (demux),
-        gst_message_new_tag (GST_OBJECT (demux), tags));
-  }
-
   /* transform some useful info to GstClockTime and remember */
   demux->start_time = gst_util_uint64_scale_int (demux->context->start_time,
       GST_SECOND, AV_TIME_BASE);
@@ -1188,6 +1192,8 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   demux->opened = TRUE;
   event = demux->seek_event;
   demux->seek_event = NULL;
+  cached_events = demux->cached_events;
+  demux->cached_events = NULL;
   GST_OBJECT_UNLOCK (demux);
 
   if (event) {
@@ -1200,32 +1206,62 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
             demux->segment.start, demux->segment.stop, demux->segment.time));
   }
 
+  while (cached_events) {
+    event = cached_events->data;
+    GST_INFO_OBJECT (demux, "pushing cached %s event: %" GST_PTR_FORMAT,
+        GST_EVENT_TYPE_NAME (event), event->structure);
+    gst_ffmpegdemux_push_event (demux, event);
+    cached_events = g_list_delete_link (cached_events, cached_events);
+  }
+
+  /* grab the global tags */
+  tags = gst_ffmpegdemux_read_tags (demux);
+  if (tags) {
+    GST_INFO_OBJECT (demux, "global tags: %" GST_PTR_FORMAT, tags);
+    gst_element_found_tags (GST_ELEMENT (demux), tags);
+  }
+
+  /* now handle the stream tags */
+  for (i = 0; i < n_streams; i++) {
+    GstFFStream *stream;
+
+    stream = gst_ffmpegdemux_get_stream (demux, demux->context->streams[i]);
+    if (stream->tags != NULL && stream->pad != NULL) {
+      GST_INFO_OBJECT (stream->pad, "stream tags: %" GST_PTR_FORMAT,
+          stream->tags);
+      gst_element_found_tags_for_pad (GST_ELEMENT (demux), stream->pad,
+          gst_tag_list_copy (stream->tags));
+    }
+  }
+
   return TRUE;
 
   /* ERRORS */
 open_failed:
   {
     GST_ELEMENT_ERROR (demux, LIBRARY, FAILED, (NULL),
-        (gst_ffmpegdemux_averror (res)));
+        ("%s", gst_ffmpegdemux_averror (res)));
     return FALSE;
   }
 no_info:
   {
     GST_ELEMENT_ERROR (demux, LIBRARY, FAILED, (NULL),
-        (gst_ffmpegdemux_averror (res)));
+        ("%s", gst_ffmpegdemux_averror (res)));
     return FALSE;
   }
 }
 
 #define GST_FFMPEG_TYPE_FIND_SIZE 4096
+#define GST_FFMPEG_TYPE_FIND_MIN_SIZE 256
+
 static void
 gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
 {
   guint8 *data;
-  GstFFMpegDemuxClassParams *params = (GstFFMpegDemuxClassParams *) priv;
-  AVInputFormat *in_plugin = params->in_plugin;
+  AVInputFormat *in_plugin = (AVInputFormat *) priv;
   gint res = 0;
   guint64 length;
+  GstCaps *sinkcaps;
 
   /* We want GST_FFMPEG_TYPE_FIND_SIZE bytes, but if the file is shorter than
    * that we'll give it a try... */
@@ -1233,6 +1269,16 @@ gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
   if (length == 0 || length > GST_FFMPEG_TYPE_FIND_SIZE)
     length = GST_FFMPEG_TYPE_FIND_SIZE;
 
+  /* The ffmpeg typefinders assume there's a certain minimum amount of data
+   * and will happily do invalid memory access if there isn't, so let's just
+   * skip the ffmpeg typefinders if the data available is too short
+   * (in which case it's unlikely to be a media file anyway) */
+  if (length < GST_FFMPEG_TYPE_FIND_MIN_SIZE) {
+    GST_LOG ("not typefinding %" G_GUINT64_FORMAT " bytes, too short", length);
+    return;
+  }
+
+  GST_LOG ("typefinding %" G_GUINT64_FORMAT " bytes", length);
   if (in_plugin->read_probe &&
       (data = gst_type_find_peek (tf, 0, length)) != NULL) {
     AVProbeData probe_data;
@@ -1251,10 +1297,12 @@ gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
       if (g_str_has_prefix (in_plugin->name, "mpegts"))
         res = MIN (res, GST_TYPE_FIND_POSSIBLE);
 
-      GST_LOG ("ffmpeg typefinder '%s' suggests %" GST_PTR_FORMAT ", p=%u%%",
-          in_plugin->name, params->sinkcaps, res);
+      sinkcaps = gst_ffmpeg_formatid_to_caps (in_plugin->name);
 
-      gst_type_find_suggest (tf, res, params->sinkcaps);
+      GST_LOG ("ffmpeg typefinder '%s' suggests %" GST_PTR_FORMAT ", p=%u%%",
+          in_plugin->name, sinkcaps, res);
+
+      gst_type_find_suggest (tf, res, sinkcaps);
     }
   }
 }
@@ -1314,8 +1362,8 @@ gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
   GST_DEBUG_OBJECT (demux,
       "pkt pts:%" GST_TIME_FORMAT
       " / size:%d / stream_index:%d / flags:%d / duration:%" GST_TIME_FORMAT
-      " / pos:%lld", GST_TIME_ARGS (timestamp), pkt.size, pkt.stream_index,
-      pkt.flags, GST_TIME_ARGS (duration), pkt.pos);
+      " / pos:%" G_GINT64_FORMAT, GST_TIME_ARGS (timestamp), pkt.size,
+      pkt.stream_index, pkt.flags, GST_TIME_ARGS (duration), (gint64) pkt.pos);
 
   /* check start_time */
   if (demux->start_time != -1 && demux->start_time > timestamp)
@@ -1410,7 +1458,7 @@ gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
 
 done:
   /* can destroy the packet now */
-  pkt.destruct (&pkt);
+  av_free_packet (&pkt);
 
   return;
 
@@ -1488,7 +1536,7 @@ drop:
     GST_DEBUG_OBJECT (demux, "dropping buffer out of segment, stream eos");
     stream->eos = TRUE;
     if (gst_ffmpegdemux_is_eos (demux)) {
-      pkt.destruct (&pkt);
+      av_free_packet (&pkt);
       GST_DEBUG_OBJECT (demux, "we are eos");
       ret = GST_FLOW_UNEXPECTED;
       goto pause;
@@ -1499,7 +1547,7 @@ drop:
   }
 no_buffer:
   {
-    pkt.destruct (&pkt);
+    av_free_packet (&pkt);
     goto pause;
   }
 }
@@ -1515,7 +1563,8 @@ gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstEvent * event)
   demux = (GstFFMpegDemux *) (GST_PAD_PARENT (sinkpad));
   ffpipe = &(demux->ffpipe);
 
-  GST_DEBUG_OBJECT (demux, "event %s", GST_EVENT_TYPE_NAME (event));
+  GST_LOG_OBJECT (demux, "%s event: %" GST_PTR_FORMAT,
+      GST_EVENT_TYPE_NAME (event), event->structure);
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_FLUSH_START:
@@ -1537,6 +1586,11 @@ gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstEvent * event)
       /* forward event */
       gst_pad_event_default (sinkpad, event);
 
+      GST_OBJECT_LOCK (demux);
+      g_list_foreach (demux->cached_events, (GFunc) gst_mini_object_unref,
+          NULL);
+      g_list_free (demux->cached_events);
+      GST_OBJECT_UNLOCK (demux);
       GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
       gst_adapter_clear (ffpipe->adapter);
       ffpipe->srcresult = GST_FLOW_OK;
@@ -1563,11 +1617,19 @@ gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstEvent * event)
        *
        * If the demuxer isn't opened, push straight away, since we'll
        * be waiting against a cond that will never be signalled. */
-      if (GST_EVENT_IS_SERIALIZED (event) && demux->opened) {
-        GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
-        while (!ffpipe->needed)
-          GST_FFMPEG_PIPE_WAIT (ffpipe);
-        GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+      if (GST_EVENT_IS_SERIALIZED (event)) {
+        if (demux->opened) {
+          GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+          while (!ffpipe->needed)
+            GST_FFMPEG_PIPE_WAIT (ffpipe);
+          GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+        } else {
+          /* queue events and send them later (esp. tag events) */
+          GST_OBJECT_LOCK (demux);
+          demux->cached_events = g_list_append (demux->cached_events, event);
+          GST_OBJECT_UNLOCK (demux);
+          goto done;
+        }
       }
       break;
   }
@@ -1758,6 +1820,10 @@ gst_ffmpegdemux_change_state (GstElement * element, GstStateChange transition)
     case GST_STATE_CHANGE_PAUSED_TO_READY:
       gst_ffmpegdemux_close (demux);
       gst_adapter_clear (demux->ffpipe.adapter);
+      g_list_foreach (demux->cached_events, (GFunc) gst_mini_object_unref,
+          NULL);
+      g_list_free (demux->cached_events);
+      demux->cached_events = NULL;
       break;
     default:
       break;
@@ -1771,7 +1837,6 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 {
   GType type;
   AVInputFormat *in_plugin;
-  GstFFMpegDemuxClassParams *params;
   gchar **extensions;
   GTypeInfo typeinfo = {
     sizeof (GstFFMpegDemuxClass),
@@ -1785,14 +1850,13 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     (GInstanceInitFunc) gst_ffmpegdemux_init,
   };
 
-  in_plugin = first_iformat;
+  in_plugin = av_iformat_next (NULL);
 
   GST_LOG ("Registering demuxers");
 
   while (in_plugin) {
     gchar *type_name, *typefind_name;
     gchar *p, *name = NULL;
-    GstCaps *sinkcaps, *audiosrccaps, *videosrccaps;
     gint rank;
     gboolean register_typefind_func = TRUE;
 
@@ -1827,7 +1891,8 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 
     /* these don't do what one would expect or
      * are only partially functional/useful */
-    if (!strcmp (in_plugin->name, "aac") || !strcmp (in_plugin->name, "wv"))
+    if (!strcmp (in_plugin->name, "aac") ||
+        !strcmp (in_plugin->name, "wv") || !strcmp (in_plugin->name, "ass"))
       goto next;
 
     /* Don't use the typefind functions of formats for which we already have
@@ -1907,41 +1972,20 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
       p++;
     }
 
-    /* Try to find the caps that belongs here */
-    sinkcaps = gst_ffmpeg_formatid_to_caps (name);
-    if (!sinkcaps) {
-      GST_WARNING ("Couldn't get sinkcaps for demuxer %s", in_plugin->name);
-      goto next;
-    }
-    /* This is a bit ugly, but we just take all formats
-     * for the pad template. We'll get an exact match
-     * when we open the stream */
-    audiosrccaps = gst_caps_new_any ();
-    videosrccaps = gst_caps_new_any ();
-
     /* construct the type */
     type_name = g_strdup_printf ("ffdemux_%s", name);
 
     /* if it's already registered, drop it */
     if (g_type_from_name (type_name)) {
-      gst_caps_unref (videosrccaps);
-      gst_caps_unref (audiosrccaps);
       g_free (type_name);
       goto next;
     }
 
     typefind_name = g_strdup_printf ("fftype_%s", name);
 
-    /* create a cache for these properties */
-    params = g_new0 (GstFFMpegDemuxClassParams, 1);
-    params->in_plugin = in_plugin;
-    params->sinkcaps = sinkcaps;
-    params->videosrccaps = videosrccaps;
-    params->audiosrccaps = audiosrccaps;
-
     /* create the type now */
     type = g_type_register_static (GST_TYPE_ELEMENT, type_name, &typeinfo, 0);
-    g_type_set_qdata (type, GST_FFDEMUX_PARAMS_QDATA, (gpointer) params);
+    g_type_set_qdata (type, GST_FFDEMUX_PARAMS_QDATA, (gpointer) in_plugin);
 
     if (in_plugin->extensions)
       extensions = g_strsplit (in_plugin->extensions, " ", 0);
@@ -1951,7 +1995,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     if (!gst_element_register (plugin, type_name, rank, type) ||
         (register_typefind_func == TRUE &&
             !gst_type_find_register (plugin, typefind_name, rank,
-                gst_ffmpegdemux_type_find, extensions, sinkcaps, params,
+                gst_ffmpegdemux_type_find, extensions, NULL, in_plugin,
                 NULL))) {
       g_warning ("Register of type ffdemux_%s failed", name);
       g_free (type_name);
@@ -1966,7 +2010,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 
   next:
     g_free (name);
-    in_plugin = in_plugin->next;
+    in_plugin = av_iformat_next (in_plugin);
   }
 
   GST_LOG ("Finished registering demuxers");
